#include "mmapGpio.h"
#include <stdio.h>
#include <unistd.h>
//#include <time.h>
#include <iostream>

#include <sys/ioctl.h>
#include <linux/fb.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>

const char *FB_NAME = "/dev/fb0";
void*   m_FrameBuffer;
struct  fb_fix_screeninfo m_FixInfo;
struct  fb_var_screeninfo m_VarInfo;
int m_FBFD;

using namespace std;

#define GPIO_VCLK	17
#define GPIO_HSYNC	27
#define GPIO_VSYNC	22
#define GPIO_PD0	18
#define GPIO_PD1	23

#define VGA_H	640
#define VGA_V	400

// Display buffer
int sBuffer[400][640];
int* iFrPtr;

struct timespec tClk;

// See: http://www.hertaville.com/rpimmapgpio.html
mmapGpio rpiGpio;

// See: https://stackoverflow.com/questions/9296059/read-pixel-value-in-bmp-file
unsigned char* readBMP(char* filename)
{
    int i;
    FILE* f = fopen(filename, "rb");
    unsigned char info[54];
    fread(info, sizeof(unsigned char), 54, f);

    int width = *(int*)&info[18];
    int height = *(int*)&info[22];
	int bpp = *(int*)&info[28];

    int size = 3 * width * height;
    unsigned char* data = new unsigned char[size];
    fread(data, sizeof(unsigned char), size, f);
    fclose(f);

	cout << "Image size: " << width << " x " << height << " x " << bpp << "\n";

	unsigned char* data_c = new unsigned char[size/3];

    for(i = 0; i < (size/3); i += 1)
    {
	data_c[i] = (data[i*3] + data[i*3+1] + data[i*3+2]);
	if (data[i] > 0)
		data[i] = 1;
    }

    return data_c;
}

// Send one VGA pixel clock cycle.
inline void vga_clock( void ) {
	rpiGpio.writePinHigh(GPIO_VCLK);

	rpiGpio.writePinLow(GPIO_VCLK);

}

// Send HSYNC signal with clocks.
inline void vga_hsync( void ) {
	// Give HSYNC signal	
	rpiGpio.writePinLow(GPIO_HSYNC);

	vga_clock();
	vga_clock();
	vga_clock();
	vga_clock();

	vga_clock();
	vga_clock();
	vga_clock();
	vga_clock();
	
	rpiGpio.writePinHigh(GPIO_HSYNC);	
}

// Write first display line with VSYNC.
inline void vga_writeLineVS( void ) {
	unsigned int iPx = 0;
	
	// Write active row in two-pixel mode
	for (iPx = 0; iPx < 320; iPx+=8) {
		// Multiple pixels in one cycle to reduce loop overhead.
		rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
		rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
		vga_clock();
		rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
		rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
		vga_clock();
		rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
		rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
		vga_clock();
		rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
		rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
		vga_clock();
	}
	
	rpiGpio.writePinHigh(GPIO_VSYNC);
	
	for (iPx = 320; iPx < 328; iPx+=8) {
		rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
	}

	rpiGpio.writePinLow(GPIO_VSYNC);

	for (iPx = 328; iPx < 640; iPx+=8) {
		rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
	}

	vga_hsync();
}

// Send blank line at end of frame.
inline void vga_blank( void ) {
	rpiGpio.writePinLow(GPIO_PD1);
	rpiGpio.writePinLow(GPIO_PD0);
	for (unsigned int iPx = 0; iPx < 640; iPx += 16) {
		vga_clock();
		vga_clock();
		vga_clock();
		vga_clock();
		vga_clock();
		vga_clock();
		vga_clock();
		vga_clock();
	}
	vga_hsync();
}

// Write single display line from buffer, no VSYNC.
inline void vga_writeLine( void ) {

	// Write active row in two-pixel mode
	// Write 2x16 pixels consecutively to speed up
	for (unsigned int iPx = 0; iPx < 640; iPx+=16) {
		rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();

		rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
                rpiGpio.writePinState(GPIO_PD1, *iFrPtr++);
                rpiGpio.writePinState(GPIO_PD0, *iFrPtr++);
                vga_clock();
	}
	vga_hsync();
}

// Write full VGA frame.
inline void writeVGA( void ) {
	
	// Start with pixel line including vertical sync, close with hsync
	iFrPtr = &sBuffer[0][0];

	vga_writeLineVS();
	
	// Continue with N-1 normal pixel lines, each with hsync
	for(unsigned int iLine = 1; iLine < 400; iLine+=16) {
		// Multiple cycles to reduce loop overhead.
		vga_writeLine();
		vga_writeLine();
		vga_writeLine();
		vga_writeLine();

		vga_writeLine();
		vga_writeLine();
		vga_writeLine();
		vga_writeLine();

		vga_writeLine();
		vga_writeLine();
		vga_writeLine();
		vga_writeLine();

		vga_writeLine();
		vga_writeLine();
		vga_writeLine();
		vga_writeLine();

	}
	
	// Last line of frame is blanked, terminate with hsync
	vga_blank();
}

// Initialize test pattern in buffer.
void initBuffer( void ) {
	unsigned int ka, kb;
	
	for(unsigned int iLine = 0; iLine < 400; iLine++) {
		for (unsigned int iCol = 0; iCol < 640; iCol++) {
			// Checkerboard of 2px fields
			ka = iLine % 2;
			kb = iCol % 2;
			//if ((ka >= 0 && ka <= 3) && (kb >= 0 && kb <= 3)) {
			if (kb==ka) {
				sBuffer[iLine][iCol] = true;
			}
		}
	}
}

int initFB( void ) {
    int iFrameBufferSize;

    m_FBFD = open(FB_NAME, O_RDWR);
    if (m_FBFD < 0) {
    	printf("Unable to open %s.\n", FB_NAME);
    	return 1;
    }

    if (ioctl(m_FBFD, FBIOGET_FSCREENINFO, &m_FixInfo) < 0) {
    	close(m_FBFD);
    	return 1;
    }

    if (ioctl(m_FBFD, FBIOGET_VSCREENINFO, &m_VarInfo) < 0) {
    	close(m_FBFD);
    	return 1;
    }

    iFrameBufferSize = m_FixInfo.line_length * m_VarInfo.yres;
    m_FrameBuffer = mmap(NULL, iFrameBufferSize, PROT_READ | PROT_WRITE,
    				 MAP_SHARED, m_FBFD,0);
    if (m_FrameBuffer == NULL) {
    	close(m_FBFD);
    	return 1;
    }

    return 0;
}


inline void updateFB( void ){
	unsigned int* p = (unsigned int*) m_FrameBuffer;

        for (unsigned int iy=0; iy<400; iy++) {
                for (unsigned int ix=0; ix<640; ix++) {
                        if ((*p++ & 0x00FFFFFF) > 0) {
                                sBuffer[iy][ix] = 1;
                        }
                        else {
                                sBuffer[iy][ix] = 0;
                        }
                }
        }
}


// Main function
int main(int argc, char* argv[]){

	int opmode;

	if (argc == 2) {
		if (string(argv[1]) == "--framebuffer")
			opmode = 2;
		else if (string(argv[1]) == "--pattern")
			opmode = 1;
		else if (string(argv[1]) == "--bitmap")
			opmode = 0;
	}
	else {
		return 1;
	}

	cout << "EL640.400 TFEL DISPLAY DRIVER\n\n";
	cout << "Initializing IO...\n";

	rpiGpio.setPinDir(GPIO_VCLK,mmapGpio::OUTPUT);
	rpiGpio.setPinDir(GPIO_HSYNC,mmapGpio::OUTPUT);
	rpiGpio.setPinDir(GPIO_VSYNC,mmapGpio::OUTPUT);
 	rpiGpio.setPinDir(GPIO_PD0,mmapGpio::OUTPUT);
	rpiGpio.setPinDir(GPIO_PD1,mmapGpio::OUTPUT);	// set GPIO4 to output

	rpiGpio.writePinHigh(GPIO_HSYNC);
	tClk.tv_sec = 0;
	tClk.tv_nsec = 400;

	if (opmode == 0) {
		cout << "Loading bitmap...\n";
		char filename[32] = "logo.bmp";
		unsigned char* bitmap = readBMP((char*)filename);

 		// Copy buffer
        	unsigned char* bmprun = bitmap;
		for(unsigned int iy=399; iy>0; iy--) {
                	for (unsigned int ix=0; ix<640; ix++) {
                        	if (*bmprun++ > 0)
                                	sBuffer[iy][ix] = 1;
                        	else
                                	sBuffer[iy][ix] = 0;
                	}
        	}
	}
	else if (opmode == 1) {
		initBuffer();
	}
	else if (opmode == 2) {
		initFB();
		updateFB();
	}

	cout << "Starting GPIO output...\n";

	while(1) {
		writeVGA();
		if (opmode == 2)
			updateFB();
		usleep(10);
	}

	return 0;
}
